/*
 * Xml.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using System.Text;
using Eagle._Attributes;
using Eagle._Components.Private;
using Eagle._Components.Public;
using Eagle._Constants;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;

namespace Eagle._Commands
{
    [ObjectId("ab8802bd-bcfd-4042-8e75-ea85c4a67959")]
    [CommandFlags(CommandFlags.Unsafe)]
    [ObjectGroup("managedEnvironment")]
    internal sealed class Xml : Core
    {
        public Xml(
            ICommandData commandData
            )
            : base(commandData)
        {
            // do nothing.
        }

        #region IEnsemble Members
        //
        // TODO: In the future, there may be a LOT more functionality here 
        //       because this is intended to be the central clearinghouse 
        //       for all Xml related commands.
        //
        private static readonly EnsembleDictionary subCommands = new EnsembleDictionary(new string[] { 
            "deserialize", "serialize"
        });

        public override EnsembleDictionary SubCommands
        {
            get { return subCommands; }
        }
        #endregion

        #region IExecute Members
        public override ReturnCode Execute(
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            ReturnCode code = ReturnCode.Ok;

            if (interpreter != null)
            {
                if (arguments != null)
                {
                    if (arguments.Count >= 2)
                    {
                        string subCommand = arguments[1];
                        bool tried = false;

                        code = RuntimeOps.TryExecuteSubCommandFromEnsemble(
                            this, interpreter, clientData, arguments, true,
                            false, ref subCommand, ref tried, ref result);

                        if ((code == ReturnCode.Ok) && !tried)
                        {
                            switch (subCommand)
                            {
                                case "deserialize":
                                    {
                                        if (arguments.Count >= 4)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] {
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-objectname", null),
                                                new Option(null, OptionFlags.MustHaveTypeValue, Index.Invalid, Index.Invalid, "-type", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nocreate", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nodispose", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-tostring", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-stricttype", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-verbose", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nocase", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-alias", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-aliasall", null),
#if NATIVE && TCL
                                                new Option(null, OptionFlags.MustHaveTclInterpreterValue, Index.Invalid, Index.Invalid, "-tcl", null),
#endif
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-noforcedelete", null),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-encoding", null),
                                                new Option(typeof(ObjectFlags), OptionFlags.MustHaveEnumValue, Index.Invalid, Index.Invalid, "-objectflags",
                                                    new Variant(ObjectFlags.Default)),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, true, ref argumentIndex, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                if ((argumentIndex != Index.Invalid) && ((argumentIndex + 2) == arguments.Count))
                                                {
#if XML && SERIALIZATION
                                                    bool create = true;

                                                    if (options.IsPresent("-nocreate"))
                                                        create = false;

                                                    bool dispose = true;

                                                    if (options.IsPresent("-nodispose"))
                                                        dispose = false;

                                                    bool toString = false;

                                                    if (options.IsPresent("-tostring"))
                                                        toString = true;

                                                    bool noCase = false;

                                                    if (options.IsPresent("-nocase"))
                                                        noCase = true;

                                                    bool strictType = false;

                                                    if (options.IsPresent("-stricttype"))
                                                        strictType = true;

                                                    bool verbose = false;

                                                    if (options.IsPresent("-verbose"))
                                                        verbose = true;

                                                    bool alias = false;

                                                    if (options.IsPresent("-alias"))
                                                        alias = true;

                                                    bool aliasAll = false;

                                                    if (options.IsPresent("-aliasall"))
                                                        aliasAll = true;

                                                    Variant value = null;
                                                    Type type = null;

                                                    if (options.IsPresent("-type", ref value))
                                                        type = (Type)value.Value;

                                                    string encodingName = null;

                                                    if (options.IsPresent("-encoding", ref value))
                                                        encodingName = value.ToString();

                                                    string interpName = null;

#if NATIVE && TCL
                                                    if (options.IsPresent("-tcl", ref value))
                                                        interpName = value.ToString();
#endif

                                                    ObjectFlags objectFlags = ObjectFlags.Default;

                                                    if (options.IsPresent("-objectflags", ref value))
                                                        objectFlags = (ObjectFlags)value.Value;

                                                    if (options.IsPresent("-noforcedelete"))
                                                        objectFlags &= ~ObjectFlags.ForceDelete;

                                                    string objectName = null;

                                                    if (options.IsPresent("-objectname", ref value))
                                                        objectName = value.ToString();

                                                    Encoding encoding = null;

                                                    if (encodingName != null)
                                                        code = Interpreter.GetEncoding(interpreter, encodingName, ref encoding, ref result);

                                                    if (code == ReturnCode.Ok)
                                                    {
                                                        //
                                                        // NOTE: If no encoding was specified, use the 
                                                        //       typical default for XML, which is UTF8.
                                                        //
                                                        if (encoding == null)
                                                            encoding = Interpreter.DefaultEncoding;

                                                        Type objectType = null;
                                                        ResultList errors = null;

                                                        code = Value.GetType(
                                                            interpreter, arguments[argumentIndex], interpreter.GetAppDomain(),
                                                            strictType, verbose, noCase, interpreter.CultureInfo, ref objectType,
                                                            ref errors);

                                                        if (code == ReturnCode.Ok)
                                                        {
                                                            byte[] bytes = encoding.GetBytes(arguments[argumentIndex + 1]);
                                                            object @object = null;

                                                            code = XmlOps.Deserialize(objectType, bytes, ref @object, ref result);

                                                            if (code == ReturnCode.Ok)
                                                                code = MarshalOps.FixupReturnValue(
                                                                    interpreter, interpreter.Binder, interpreter.CultureInfo,
                                                                    type, objectFlags, RuntimeOps.GetInvokeOptions(aliasAll),
                                                                    objectName, interpName, @object, create, dispose, alias,
                                                                    aliasAll, toString, ref result);
                                                        }
                                                        else
                                                        {
                                                            errors.Insert(0, String.Format(
                                                                "type \"{0}\" not found",
                                                                arguments[argumentIndex]));

                                                            result = errors;
                                                        }
                                                    }
#else
                                                    result = "not implemented";
                                                    code = ReturnCode.Error;
#endif
                                                }
                                                else
                                                {
                                                    if ((argumentIndex != Index.Invalid) &&
                                                        Option.LooksLikeOption(arguments[argumentIndex]))
                                                    {
                                                        result = OptionDictionary.BadOption(options, arguments[argumentIndex]);
                                                    }
                                                    else
                                                    {
                                                        result = "wrong # args: should be \"xml deserialize ?options? type xml\"";
                                                    }

                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"xml deserialize ?options? type xml\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "serialize":
                                    {
                                        if (arguments.Count >= 4)
                                        {
                                            OptionDictionary options = new OptionDictionary(
                                                new Option[] { 
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-stricttype", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-verbose", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, "-nocase", null),
                                                new Option(null, OptionFlags.MustHaveValue, Index.Invalid, Index.Invalid, "-encoding", null),
                                                new Option(null, OptionFlags.None, Index.Invalid, Index.Invalid, Option.EndOfOptions, null)
                                            });

                                            int argumentIndex = Index.Invalid;

                                            code = interpreter.GetOptions(options, arguments, 0, 2, Index.Invalid, true, ref argumentIndex, ref result);

                                            if (code == ReturnCode.Ok)
                                            {
                                                if ((argumentIndex != Index.Invalid) && ((argumentIndex + 2) == arguments.Count))
                                                {
#if XML && SERIALIZATION
                                                    bool noCase = false;

                                                    if (options.IsPresent("-nocase"))
                                                        noCase = true;

                                                    bool strictType = false;

                                                    if (options.IsPresent("-stricttype"))
                                                        strictType = true;

                                                    bool verbose = false;

                                                    if (options.IsPresent("-verbose"))
                                                        verbose = true;

                                                    Variant value = null;
                                                    string encodingName = null;

                                                    if (options.IsPresent("-encoding", ref value))
                                                        encodingName = value.ToString();

                                                    Encoding encoding = null;

                                                    if (encodingName != null)
                                                        code = Interpreter.GetEncoding(interpreter, encodingName, ref encoding, ref result);

                                                    if (code == ReturnCode.Ok)
                                                    {
                                                        //
                                                        // NOTE: If no encoding was specified, use the 
                                                        //       typical default for XML, which is UTF8.
                                                        //
                                                        if (encoding == null)
                                                            encoding = Interpreter.DefaultEncoding; 

                                                        Type type = null;
                                                        ResultList errors = null;

                                                        code = Value.GetType(
                                                            interpreter, arguments[argumentIndex], interpreter.GetAppDomain(),
                                                            strictType, verbose, noCase, interpreter.CultureInfo, ref type, ref errors);

                                                        if (code == ReturnCode.Ok)
                                                        {
                                                            IObject @object = null;

                                                            code = interpreter.GetObject(arguments[argumentIndex + 1], true, true, 
                                                                ref @object, ref result);

                                                            if (code == ReturnCode.Ok)
                                                            {
                                                                byte[] bytes = null;

                                                                code = XmlOps.Serialize((@object != null) ? @object.Value : null,
                                                                    type, ref bytes, ref result);

                                                                if (code == ReturnCode.Ok)
                                                                    result = encoding.GetString(bytes);
                                                            }
                                                        }
                                                        else
                                                        {
                                                            errors.Insert(0, String.Format(
                                                                "type \"{0}\" not found",
                                                                arguments[argumentIndex]));

                                                            result = errors;
                                                        }
                                                    }
#else
                                                    result = "not implemented";
                                                    code = ReturnCode.Error;
#endif
                                                }
                                                else
                                                {
                                                    if ((argumentIndex != Index.Invalid) &&
                                                        Option.LooksLikeOption(arguments[argumentIndex]))
                                                    {
                                                        result = OptionDictionary.BadOption(options, arguments[argumentIndex]);
                                                    }
                                                    else
                                                    {
                                                        result = "wrong # args: should be \"xml serialize ?options? type object\"";
                                                    }

                                                    code = ReturnCode.Error;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"xml serialize ?options? type object\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                default:
                                    {
                                        result = RuntimeOps.BadSubCommand(null, null, subCommand, this, null, null);
                                        code = ReturnCode.Error;
                                        break;
                                    }
                            }
                        }
                    }
                    else
                    {
                        result = "wrong # args: should be \"xml option ?arg ...?\"";
                        code = ReturnCode.Error;
                    }
                }
                else
                {
                    result = "invalid argument list";
                    code = ReturnCode.Error;
                }
            }
            else
            {
                result = "invalid interpreter";
                code = ReturnCode.Error;
            }

            return code;
        }
        #endregion
    }
}
