/*
 * Debugger.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using Eagle._Attributes;
using Eagle._Components.Public;
using Eagle._Components.Public.Delegates;

#if BREAKPOINTS
using Eagle._Containers.Private;
#endif

using Eagle._Containers.Public;
using Eagle._Interfaces.Private;
using Eagle._Interfaces.Public;

namespace Eagle._Components.Private
{
    [ObjectId("9be2b241-bee5-428f-9df8-df354ef63ea2")]
    internal sealed class Debugger :
#if ISOLATED_INTERPRETERS || ISOLATED_PLUGINS
        ScriptMarshalByRefObject,
#endif
        IDebugger, IDisposable
    {
        #region Public Constants
        public static readonly string Name = typeof(Debugger).Name;

#if SHELL
        public static readonly EngineFlags BreakEngineFlags = EngineFlags.None;
#endif

        public static readonly EngineFlags WatchEngineFlags = EngineFlags.None;

        public static readonly EngineFlags InteractiveEngineFlags =
            EngineFlags.NoBreakpoint | EngineFlags.Interactive;

        public static readonly bool DefaultEnabled = true;

        public static readonly BreakpointType DefaultTypes = BreakpointType.Express;
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Enumerations
        [ObjectId("b53c003a-bb94-4c10-900c-7403dc36c2a8")]
        private enum Context
        {
            CurrentIndex = 0,          // current value of property.
            SaveIndex = 1,             // saved value of property (internal save/restore semantics).
            FirstIndex = CurrentIndex,
            LastIndex = SaveIndex
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Public Constructors
        public Debugger(
            bool isolated,
            string culture,
            CreateFlags createFlags,
            AppDomain appDomain,
            IHost host
            )
        {
            if (isolated)
                interpreter = CreateInterpreter(culture, createFlags, appDomain, host);
            else
                interpreter = null;

            ReturnCode code;
            Result error = null;

            code = Initialize(ref error);

            if (code != ReturnCode.Ok)
                DebugOps.Complain(code, error);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Static "Factory" Members
        public static Interpreter CreateInterpreter(
            string culture,
            CreateFlags createFlags,
            AppDomain appDomain,
            IHost host
            )
        {
            //
            // NOTE: First, mask off flags that we know to be invalid for
            //       debugger interpreters.  Next, add a flag to prevent a
            //       default interpreter host from being created.  Finally,
            //       create an isolated debugging interpreter with the right
            //       set of options.  This is actually a very delicate
            //       operation because we need to clone the provided host into
            //       the new interpreter; however, we cannot do that until it
            //       is created.  The solution is to create the interpreter
            //       without any host whatsoever and then use it to create a
            //       clone of the provided host.  The newly cloned host can
            //       then be assigned to the newly created interpreter.
            //
            createFlags &= ~CreateFlags.NonDebuggerUse;
            createFlags |= CreateFlags.NoHost;

            Interpreter interpreter = Interpreter.Create(culture, createFlags, appDomain);

            if (interpreter != null)
                interpreter.Host = (host != null) ? host.Clone(interpreter) : null;

            return interpreter;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Static Watch / Break Methods
        public static ReturnCode Watch(
            IDebugger debugger,
            Interpreter interpreter,
            ReturnCode code,
            BreakpointType breakpointType,
            string breakpointName,
            IToken token,
            ITraceInfo traceInfo,
            EngineFlags engineFlags,
            SubstitutionFlags substitutionFlags,
            EventFlags eventFlags,
            ExpressionFlags expressionFlags,
            HeaderFlags headerFlags,
            ref Result result
            )
        {
            bool exit = false;

            return Watch(
                debugger, interpreter, code, breakpointType, breakpointName,
                token, traceInfo, engineFlags, substitutionFlags, eventFlags,
                expressionFlags, headerFlags, ref exit, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode Watch(
            IDebugger debugger,
            Interpreter interpreter,
            ReturnCode code,
            BreakpointType breakpointType,
            string breakpointName,
            IToken token,
            ITraceInfo traceInfo,
            EngineFlags engineFlags,
            SubstitutionFlags substitutionFlags,
            EventFlags eventFlags,
            ExpressionFlags expressionFlags,
            HeaderFlags headerFlags,
            ref bool exit,
            ref Result result
            )
        {
            return Break(
                debugger, interpreter, code, breakpointType, breakpointName,
                token, traceInfo, engineFlags | WatchEngineFlags, substitutionFlags,
                eventFlags, expressionFlags, headerFlags, null, null, ref exit,
                ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode Break(
            IDebugger debugger,
            Interpreter interpreter,
            ReturnCode code,
            BreakpointType breakpointType,
            string breakpointName,
            IToken token,
            ITraceInfo traceInfo,
            EngineFlags engineFlags,
            SubstitutionFlags substitutionFlags,
            EventFlags eventFlags,
            ExpressionFlags expressionFlags,
            HeaderFlags headerFlags,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            bool exit = false;

            return Break(
                debugger, interpreter, code, breakpointType, breakpointName,
                token, traceInfo, engineFlags, substitutionFlags, eventFlags,
                expressionFlags, headerFlags, clientData, arguments, ref exit,
                ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode Break(
            IDebugger debugger,
            Interpreter interpreter,
            ReturnCode code,
            BreakpointType breakpointType,
            string breakpointName,
            IToken token,
            ITraceInfo traceInfo,
            EngineFlags engineFlags,
            SubstitutionFlags substitutionFlags,
            EventFlags eventFlags,
            ExpressionFlags expressionFlags,
            HeaderFlags headerFlags,
            IClientData clientData,
            ArgumentList arguments,
            ref bool exit,
            ref Result result
            )
        {
            if (interpreter != null)
            {
                if (interpreter.Interactive)
                {
                    InteractiveLoopCallback interactiveLoopCallback =
                        interpreter.InteractiveLoopCallback;

                    if (interactiveLoopCallback != null)
                    {
                        return interactiveLoopCallback(interpreter, true, null,
                            code, breakpointType, breakpointName, token, traceInfo,
                            engineFlags | BreakEngineFlags, substitutionFlags,
                            eventFlags, expressionFlags, headerFlags, clientData,
                            arguments, ref exit, ref result);
                    }
                    else
                    {
#if SHELL
                        //
                        // NOTE: This is the only place in this class where the
                        //       InteractiveLoop method may be called.  All other
                        //       methods in the Debugger class and/or any external
                        //       classes that desire an interactive debugging
                        //       functionality should call the Debugger.Break
                        //       method.
                        //
                        return Interpreter.InteractiveLoop(interpreter, true, null,
                            code, breakpointType, breakpointName, token, traceInfo,
                            engineFlags | BreakEngineFlags, substitutionFlags,
                            eventFlags, expressionFlags, headerFlags, clientData,
                            arguments, ref exit, ref result);
#else
                        result = "not implemented";
#endif
                    }
                }
                else
                {
                    result = "cannot break into interactive loop";
                }
            }
            else
            {
                result = "invalid interpreter";
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IGetInterpreter / ISetInterpreter Members
        private Interpreter interpreter; /* special out-of-band interpreter */
        public Interpreter Interpreter
        {
            get { CheckDisposed(); return interpreter; }
            set { CheckDisposed(); interpreter = value; }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IDebugger Members
        private int suspendCount;
        public int SuspendCount
        {
            get { CheckDisposed(); return suspendCount; }
            set { CheckDisposed(); suspendCount = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool[] enabled;
        public bool Enabled
        {
            get { CheckDisposed(); return enabled[(int)Context.CurrentIndex]; }
            set { CheckDisposed(); enabled[(int)Context.CurrentIndex] = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        //
        // NOTE: Special flag for scripts to use to detect they are being executed
        //       while in the interactive debugger.
        //
        private int[] active;
        public int Active
        {
            get { CheckDisposed(); return active[(int)Context.CurrentIndex]; }
            set { CheckDisposed(); active[(int)Context.CurrentIndex] = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool[] singleStep;
        public bool SingleStep
        {
            get { CheckDisposed(); return singleStep[(int)Context.CurrentIndex]; }
            set { CheckDisposed(); singleStep[(int)Context.CurrentIndex] = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if BREAKPOINTS
        private bool[] breakOnToken;
        public bool BreakOnToken
        {
            get { CheckDisposed(); return breakOnToken[(int)Context.CurrentIndex]; }
            set { CheckDisposed(); breakOnToken[(int)Context.CurrentIndex] = value; }
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool[] breakOnExecute;
        public bool BreakOnExecute
        {
            get { CheckDisposed(); return breakOnExecute[(int)Context.CurrentIndex]; }
            set { CheckDisposed(); breakOnExecute[(int)Context.CurrentIndex] = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool[] breakOnCancel;
        public bool BreakOnCancel
        {
            get { CheckDisposed(); return breakOnCancel[(int)Context.CurrentIndex]; }
            set { CheckDisposed(); breakOnCancel[(int)Context.CurrentIndex] = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool[] breakOnError;
        public bool BreakOnError
        {
            get { CheckDisposed(); return breakOnError[(int)Context.CurrentIndex]; }
            set { CheckDisposed(); breakOnError[(int)Context.CurrentIndex] = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool[] breakOnReturn;
        public bool BreakOnReturn
        {
            get { CheckDisposed(); return breakOnReturn[(int)Context.CurrentIndex]; }
            set { CheckDisposed(); breakOnReturn[(int)Context.CurrentIndex] = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool[] breakOnTest;
        public bool BreakOnTest
        {
            get { CheckDisposed(); return breakOnTest[(int)Context.CurrentIndex]; }
            set { CheckDisposed(); breakOnTest[(int)Context.CurrentIndex] = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool[] breakOnExit;
        public bool BreakOnExit
        {
            get { CheckDisposed(); return breakOnExit[(int)Context.CurrentIndex]; }
            set { CheckDisposed(); breakOnExit[(int)Context.CurrentIndex] = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private long[] steps;
        public long Steps
        {
            get { CheckDisposed(); return steps[(int)Context.CurrentIndex]; }
            set { CheckDisposed(); steps[(int)Context.CurrentIndex] = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private BreakpointType[] types;
        public BreakpointType Types
        {
            get { CheckDisposed(); return types[(int)Context.CurrentIndex]; }
            set { CheckDisposed(); types[(int)Context.CurrentIndex] = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if BREAKPOINTS
        private BreakpointDictionary[] breakpoints;
        public BreakpointDictionary Breakpoints
        {
            get { CheckDisposed(); return breakpoints[(int)Context.CurrentIndex]; }
            set { CheckDisposed(); breakpoints[(int)Context.CurrentIndex] = value; }
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string[] command;
        public string Command
        {
            get { CheckDisposed(); return command[(int)Context.CurrentIndex]; }
            set { CheckDisposed(); command[(int)Context.CurrentIndex] = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private Result[] result;
        public Result Result
        {
            get { CheckDisposed(); return result[(int)Context.CurrentIndex]; }
            set { CheckDisposed(); result[(int)Context.CurrentIndex] = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public int SetActive(
            bool active
            )
        {
            CheckDisposed();

            if (active)
                return ++this.active[(int)Context.CurrentIndex];
            else if (this.active[(int)Context.CurrentIndex] > 0)
                return --this.active[(int)Context.CurrentIndex];
            else
                return 0;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public long NextStep()
        {
            CheckDisposed();

            if (steps[(int)Context.CurrentIndex] > 0)
                return --steps[(int)Context.CurrentIndex];
            else
                return 0;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if BREAKPOINTS
        public ReturnCode GetBreakpointList(
            string pattern,
            bool noCase,
            ref IStringList list,
            ref Result error
            )
        {
            CheckDisposed();

            if (breakpoints[(int)Context.CurrentIndex] != null)
            {
                list = breakpoints[(int)Context.CurrentIndex].ToList(pattern, noCase);

                return ReturnCode.Ok;
            }
            else
            {
                error = "breakpoints not available";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode MatchBreakpoint(
            IScriptLocation location,
            ref bool match
            )
        {
            CheckDisposed();

            Result error = null;

            return MatchBreakpoint(location, ref match, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode MatchBreakpoint(
            IScriptLocation location,
            ref bool match,
            ref Result error
            )
        {
            CheckDisposed();

            if (location != null)
            {
                if (ScriptLocation.Check(location, false))
                {
                    string fileName = location.FileName;

                    if (fileName != null)
                        fileName = ScriptLocation.NormalizeFileName(fileName);

                    //
                    // NOTE: *WARNING: Empty file names are allowed here, please do
                    //       not change this to !String.IsNullOrEmpty.
                    //
                    if (fileName != null)
                    {
                        if (breakpoints[(int)Context.CurrentIndex] != null)
                        {
                            Breakpoint breakpoint;

                            if (breakpoints[(int)Context.CurrentIndex].TryGetValue(
                                    fileName, out breakpoint))
                            {
                                return breakpoint.Match(location, ref match, ref error);
                            }
                            else
                            {
                                match = false; // NOTE: It was not found.
                                return ReturnCode.Ok;
                            }
                        }
                        else
                        {
                            error = "breakpoints not available";
                        }
                    }
                    else
                    {
                        error = "invalid script location file name";
                    }
                }
                else
                {
                    error = "bad script location";
                }
            }
            else
            {
                error = "invalid script location";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode ClearBreakpoint(
            IScriptLocation location,
            ref bool match,
            ref Result error
            )
        {
            CheckDisposed();

            if (location != null)
            {
                if (ScriptLocation.Check(location, false))
                {
                    string fileName = location.FileName;

                    if (fileName != null)
                        fileName = ScriptLocation.NormalizeFileName(fileName);

                    //
                    // NOTE: *WARNING: Empty file names are allowed here, please do
                    //       not change this to !String.IsNullOrEmpty.
                    //
                    if (fileName != null)
                    {
                        if (breakpoints[(int)Context.CurrentIndex] != null)
                        {
                            Breakpoint breakpoint;

                            if (breakpoints[(int)Context.CurrentIndex].TryGetValue(
                                    fileName, out breakpoint))
                            {
                                return breakpoint.Clear(location, ref match, ref error);
                            }
                            else
                            {
                                match = false; // NOTE: It was not already found.
                                return ReturnCode.Ok;
                            }
                        }
                        else
                        {
                            error = "breakpoints not available";
                        }
                    }
                    else
                    {
                        error = "invalid script location file name";
                    }
                }
                else
                {
                    error = "bad script location";
                }
            }
            else
            {
                error = "invalid script location";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SetBreakpoint(
            IScriptLocation location,
            ref bool match,
            ref Result error
            )
        {
            CheckDisposed();

            if (location != null)
            {
                if (ScriptLocation.Check(location, false))
                {
                    string fileName = location.FileName;

                    if (fileName != null)
                        fileName = ScriptLocation.NormalizeFileName(fileName);

                    //
                    // NOTE: *WARNING: Empty file names are allowed here, please do
                    //       not change this to !String.IsNullOrEmpty.
                    //
                    if (fileName != null)
                    {
                        if (breakpoints[(int)Context.CurrentIndex] != null)
                        {
                            Breakpoint breakpoint;

                            if (breakpoints[(int)Context.CurrentIndex].TryGetValue(
                                    fileName, out breakpoint))
                            {
                                return breakpoint.Set(location, ref match, ref error);
                            }
                            else
                            {
                                breakpoints[(int)Context.CurrentIndex].Add(
                                    fileName, new Breakpoint(location));

                                match = false; // NOTE: It was not already found.
                                return ReturnCode.Ok;
                            }
                        }
                        else
                        {
                            error = "breakpoints not available";
                        }
                    }
                    else
                    {
                        error = "invalid script location file name";
                    }
                }
                else
                {
                    error = "bad script location";
                }
            }
            else
            {
                error = "invalid script location";
            }

            return ReturnCode.Error;
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode Initialize(
            ref Result error
            )
        {
            CheckDisposed();

            //
            // NOTE: Initializes the debugger to the its initial default state
            //       (i.e. enabled and inactive).
            //
            suspendCount = 0;

            enabled = new bool[] { DefaultEnabled, false };
            active = new int[] { 0, 0 };
            singleStep = new bool[] { false, false };

#if BREAKPOINTS
            breakOnToken = new bool[] { false, false };
#endif

            breakOnExecute = new bool[] { false, false };
            breakOnCancel = new bool[] { false, false };
            breakOnError = new bool[] { false, false };
            breakOnReturn = new bool[] { false, false };
            breakOnTest = new bool[] { false, false };
            breakOnExit = new bool[] { false, false };
            steps = new long[] { 0, 0 };
            types = new BreakpointType[] { DefaultTypes, BreakpointType.None };

#if BREAKPOINTS
            breakpoints = new BreakpointDictionary[] {
                new BreakpointDictionary(), null
            };
#endif

            command = new string[] { null, null };
            result = new Result[] { null, null };

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode Reset(
            ref Result error
            )
        {
            CheckDisposed();

            //
            // NOTE: Resets the internal state of the debugger (i.e. disabled
            //       and inactive).
            //
            suspendCount = 0;

            for (int index = (int)Context.FirstIndex; index <= (int)Context.LastIndex; index++)
            {
                enabled[index] = false;
                active[index] = 0;
                singleStep[index] = false;

#if BREAKPOINTS
                breakOnToken[index] = false;
#endif

                breakOnExecute[index] = false;
                breakOnCancel[index] = false;
                breakOnError[index] = false;
                breakOnReturn[index] = false;
                breakOnTest[index] = false;
                breakOnExit[index] = false;
                steps[index] = 0;
                types[index] = BreakpointType.None;

#if BREAKPOINTS
                breakpoints[index] = null;
#endif

                command[index] = null;
                result[index] = null;
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode Suspend(
            ref Result error
            )
        {
            CheckDisposed();

            if (++suspendCount == 1)
            {
                //
                // NOTE: Save current debugger state information.
                //
                enabled[(int)Context.SaveIndex] = enabled[(int)Context.CurrentIndex];
                active[(int)Context.SaveIndex] = active[(int)Context.CurrentIndex];
                singleStep[(int)Context.SaveIndex] = singleStep[(int)Context.CurrentIndex];

#if BREAKPOINTS
                breakOnToken[(int)Context.SaveIndex] = breakOnToken[(int)Context.CurrentIndex];
#endif

                breakOnExecute[(int)Context.SaveIndex] = breakOnExecute[(int)Context.CurrentIndex];
                breakOnCancel[(int)Context.SaveIndex] = breakOnCancel[(int)Context.CurrentIndex];
                breakOnError[(int)Context.SaveIndex] = breakOnError[(int)Context.CurrentIndex];
                breakOnReturn[(int)Context.SaveIndex] = breakOnReturn[(int)Context.CurrentIndex];
                breakOnTest[(int)Context.SaveIndex] = breakOnTest[(int)Context.CurrentIndex];
                breakOnExit[(int)Context.SaveIndex] = breakOnExit[(int)Context.CurrentIndex];
                steps[(int)Context.SaveIndex] = steps[(int)Context.CurrentIndex];
                types[(int)Context.SaveIndex] = types[(int)Context.CurrentIndex];

#if BREAKPOINTS
                breakpoints[(int)Context.SaveIndex] = breakpoints[(int)Context.CurrentIndex];
#endif

                command[(int)Context.SaveIndex] = command[(int)Context.CurrentIndex];
                result[(int)Context.SaveIndex] = result[(int)Context.CurrentIndex];

                //
                // NOTE: Reset current debugger state information.
                //
                enabled[(int)Context.CurrentIndex] = false;
                active[(int)Context.CurrentIndex] = 0;
                singleStep[(int)Context.CurrentIndex] = false;

#if BREAKPOINTS
                breakOnToken[(int)Context.CurrentIndex] = false;
#endif

                breakOnExecute[(int)Context.CurrentIndex] = false;
                breakOnCancel[(int)Context.CurrentIndex] = false;
                breakOnError[(int)Context.CurrentIndex] = false;
                breakOnReturn[(int)Context.CurrentIndex] = false;
                breakOnTest[(int)Context.CurrentIndex] = false;
                breakOnExit[(int)Context.CurrentIndex] = false;
                steps[(int)Context.CurrentIndex] = 0;
                types[(int)Context.CurrentIndex] = BreakpointType.None;

#if BREAKPOINTS
                breakpoints[(int)Context.CurrentIndex] = null;
#endif

                command[(int)Context.CurrentIndex] = null;
                result[(int)Context.CurrentIndex] = null;
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode Resume(
            ref Result error
            )
        {
            CheckDisposed();

            if (--suspendCount == 0)
            {
                //
                // NOTE: Restore current debugger state information.
                //
                enabled[(int)Context.CurrentIndex] = enabled[(int)Context.SaveIndex];
                active[(int)Context.CurrentIndex] = active[(int)Context.SaveIndex];
                singleStep[(int)Context.CurrentIndex] = singleStep[(int)Context.SaveIndex];

#if BREAKPOINTS
                breakOnToken[(int)Context.CurrentIndex] = breakOnToken[(int)Context.SaveIndex];
#endif

                breakOnExecute[(int)Context.CurrentIndex] = breakOnExecute[(int)Context.SaveIndex];
                breakOnCancel[(int)Context.CurrentIndex] = breakOnCancel[(int)Context.SaveIndex];
                breakOnError[(int)Context.CurrentIndex] = breakOnError[(int)Context.SaveIndex];
                breakOnReturn[(int)Context.CurrentIndex] = breakOnReturn[(int)Context.SaveIndex];
                breakOnTest[(int)Context.CurrentIndex] = breakOnTest[(int)Context.SaveIndex];
                breakOnExit[(int)Context.CurrentIndex] = breakOnExit[(int)Context.SaveIndex];
                steps[(int)Context.CurrentIndex] = steps[(int)Context.SaveIndex];
                types[(int)Context.CurrentIndex] = types[(int)Context.SaveIndex];

#if BREAKPOINTS
                breakpoints[(int)Context.CurrentIndex] = breakpoints[(int)Context.SaveIndex];
#endif

                command[(int)Context.CurrentIndex] = command[(int)Context.SaveIndex];
                result[(int)Context.CurrentIndex] = result[(int)Context.SaveIndex];

                //
                // NOTE: Reset saved debugger state information.
                //
                enabled[(int)Context.SaveIndex] = false;
                active[(int)Context.SaveIndex] = 0;
                singleStep[(int)Context.SaveIndex] = false;

#if BREAKPOINTS
                breakOnToken[(int)Context.SaveIndex] = false;
#endif

                breakOnExecute[(int)Context.SaveIndex] = false;
                breakOnCancel[(int)Context.SaveIndex] = false;
                breakOnError[(int)Context.SaveIndex] = false;
                breakOnReturn[(int)Context.SaveIndex] = false;
                breakOnTest[(int)Context.SaveIndex] = false;
                breakOnExit[(int)Context.SaveIndex] = false;
                steps[(int)Context.SaveIndex] = 0;
                types[(int)Context.SaveIndex] = BreakpointType.None;

#if BREAKPOINTS
                breakpoints[(int)Context.SaveIndex] = null;
#endif

                command[(int)Context.SaveIndex] = null;
                result[(int)Context.SaveIndex] = null;
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode Watch(
            Interpreter interpreter,
            ReturnCode code,
            BreakpointType breakpointType,
            string breakpointName,
            IToken token,
            ITraceInfo traceInfo,
            EngineFlags engineFlags,
            SubstitutionFlags substitutionFlags,
            EventFlags eventFlags,
            ExpressionFlags expressionFlags,
            HeaderFlags headerFlags,
            ref bool exit,
            ref Result result
            )
        {
            CheckDisposed();

            return Watch(
                this, interpreter, code, breakpointType, breakpointName,
                token, traceInfo, engineFlags, substitutionFlags, eventFlags,
                expressionFlags, headerFlags, ref exit, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode Break(
            Interpreter interpreter,
            ReturnCode code,
            BreakpointType breakpointType,
            string breakpointName,
            IToken token,
            ITraceInfo traceInfo,
            EngineFlags engineFlags,
            SubstitutionFlags substitutionFlags,
            EventFlags eventFlags,
            ExpressionFlags expressionFlags,
            HeaderFlags headerFlags,
            IClientData clientData,
            ArgumentList arguments,
            ref bool exit,
            ref Result result
            )
        {
            CheckDisposed();

            return Break(
                this, interpreter, code, breakpointType, breakpointName,
                token, traceInfo, engineFlags, substitutionFlags, eventFlags,
                expressionFlags, headerFlags, clientData, arguments, ref exit,
                ref result);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IDisposable "Pattern" Members
        private bool disposed;
        private void CheckDisposed() /* throw */
        {
#if THROW_ON_DISPOSED
            if (disposed && Engine.IsThrowOnDisposed(null, false))
                throw new ObjectDisposedException(typeof(Debugger).Name);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private /* protected virtual */ void Dispose(bool disposing)
        {
            if (!disposed)
            {
                if (disposing)
                {
                    ////////////////////////////////////
                    // dispose managed resources here...
                    ////////////////////////////////////

                    if (interpreter != null)
                    {
                        interpreter.Dispose();
                        interpreter = null;
                    }
                }

                //////////////////////////////////////
                // release unmanaged resources here...
                //////////////////////////////////////

                disposed = true;
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IDisposable Members
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Destructor
        ~Debugger()
        {
            Dispose(false);
        }
        #endregion
    }
}
