/*
 * NativePackage.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using System.Collections.Generic;
using Eagle._Attributes;
using Eagle._Components.Private;
using Eagle._Components.Private.Tcl;
using Eagle._Constants;
using Eagle._Containers.Private;
using Eagle._Containers.Public;
using Eagle._Interfaces.Private.Tcl;
using Eagle._Interfaces.Public;

namespace Eagle._Components.Public
{
    [ObjectId("2e8eae65-3e12-4eb9-8695-c871cc23a57a")]
    public static class NativePackage
    {
        #region Private Constants
        //
        // NOTE: This is the name of the native command in the Tcl interpreter
        //       that will be bridged to the Eagle interpreter (i.e. the native
        //       endpoint).
        //
        private static readonly string nativeCommandName =
            GlobalState.GetPackageName(true);

        //
        // NOTE: This is the name of the managed command in the Eagle
        //       interpreter that will be bridged to the Tcl interpreter
        //       (i.e. the managed endpoint).
        //
        private static readonly string managedCommandName =
            typeof(_Commands.Eval).Name.ToLowerInvariant();

        //
        // NOTE: These name prefixes are used when building the script-visible
        //       names for the Tcl interpreters used by this class.  Also, see
        //       the constants tclMasterInterpPrefix, tclSafeInterpPrefix, and
        //       tclInterpPrefix in the Interpreter class.
        //
        private const string tclNativeMasterInterpPrefix = "nativeMasterInterp";
        private const string tclNativeSafeMasterInterpPrefix = "nativeSafeMasterInterp";
        private const string tclNativeInterpPrefix = "nativeInterp";
        private const string tclNativeSafeInterpPrefix = "nativeSafeInterp";

        //
        // NOTE: In the argument string passed from native code, we need at
        //       least the module handle, the Tcl interpreter pointer, and the
        //       Tcl interpreter safety indicator.
        //
        private const int MinimumArgumentCount = 4;

        //
        // NOTE: The protocol Id string we expect in the argument string passed
        //       from native code.
        //
        private const string ProtocolId = "Garuda_v1.0";

        //
        // NOTE: This is the error message returned when the string argument
        //       cannot be parsed properly (i.e. not a list, not enough
        //       sub-arguments, etc).
        //
        private const string ParseArgumentError = "could not parse argument " +
            "string, expected at least [{0} <IntPtr> <IntPtr> <Boolean>]: {1}";

        //
        // NOTE: This is the error message returned when the "safe" mode of the
        //       Eagle interpreter is unsuitable for the "safe" mode of the Tcl
        //       interpreter.
        //
        private const string SafeUnsafeError = "cannot use safe Tcl " +
            "interpreter {0} with unsafe Eagle interpreter {1}";

        //
        // NOTE: This is the error message returned when the Tcl interpreter
        //       has already been attached to the bridge.
        //
        private const string AlreadyAttachedError = "cannot attach Tcl " +
            "interpreter {0} to Eagle interpreter {1}, already attached";
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Private Data
        //
        // NOTE: This object is used with the lock statement to protect
        //       access to the other static fields.
        //
        private static readonly object syncRoot = new object();

        //
        // NOTE: The Eagle interpreter holding the Tcl integration components
        //       created and used by this class.
        //
        private static Interpreter interpreter = null;

        //
        // NOTE: The list of Tcl interpreters we know originated outside the
        //       direct control of Eagle (i.e. we should not try to delete
        //       them).
        //
        private static IntPtrDictionary tclInterps = null;

        //
        // NOTE: The thread Id we were started up on by Tcl.
        //
        private static int threadId = 0;
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Internal Integration Methods
        //
        // NOTE: Used by the CheckInterp method of the TclApi class.
        //
        internal static bool FindTclInterpreterThreadId(
            IntPtr interp,
            ref int threadId
            )
        {
            lock (syncRoot)
            {
                if ((tclInterps != null) &&
                    tclInterps.ContainsValue(interp) /* O(N) */)
                {
                    int localThreadId = NativePackage.threadId;

                    if (localThreadId != 0)
                    {
                        threadId = localThreadId;
                        return true;
                    }
                }

                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////

        //
        // NOTE: Used by the DoOneEvent method of the TclWrapper class.
        //
        internal static bool IsTclInterpreterActive()
        {
            lock (syncRoot)
            {
                return ((interpreter != null) && (tclInterps != null) &&
                    (tclInterps.Count > 0));
            }
        }

        ///////////////////////////////////////////////////////////////////////

        //
        // NOTE: Used by the Execute method of the _Commands.Tcl class to help
        //       implement the [tcl master] sub-command.
        //
        internal static ReturnCode GetMasterTclInterpreter(
            ref string name,
            ref IntPtr interp,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (HasTclInterpreters(ref error))
                {
                    //
                    // HACK: Use the convention that we always prefix the
                    //       master Tcl interpreter with a specially formatted
                    //       name in order to find it (we cannot simply use the
                    //       Tcl interpreter at index zero because the Tcl
                    //       interpreters are maintained in a dictionary).
                    //
                    string pattern = FormatOps.Id(tclNativeMasterInterpPrefix,
                        null, Characters.Asterisk.ToString());

                    string safePattern = FormatOps.Id(tclNativeSafeMasterInterpPrefix,
                        null, Characters.Asterisk.ToString());

                    //
                    // NOTE: Search all the Tcl interpreters in this interpreter
                    //       for the master Tcl interp stopping as soon as we
                    //       find a valid one (there should only be one valid
                    //       master Tcl interpreter per interpreter at any given
                    //       time).
                    //
                    string key = null;
                    IntPtr value = IntPtr.Zero;

                    foreach (KeyValuePair<string, IntPtr> pair in tclInterps)
                    {
                        //
                        // NOTE: Hard-coded match mode is OK here.
                        //
                        if ((StringOps.Match(
                                MatchMode.Glob, pair.Key, pattern, false) ||
                            StringOps.Match(
                                MatchMode.Glob, pair.Key, safePattern, false)) &&
                            (pair.Value != IntPtr.Zero))
                        {
                            key = pair.Key;
                            value = pair.Value;
                            break;
                        }
                    }

                    //
                    // NOTE: Did we find the handle for the master Tcl
                    //       interpreter?
                    //
                    if (key != null)
                    {
                        name = key;
                        interp = value;

                        return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "Tcl master interpreter is not available";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        //
        // NOTE: Used by the DeleteTclInterpreter and DisposeTcl methods of
        //       the Interpreter class.
        //
        internal static bool ShouldDeleteTclInterpreter(
            string interpName,
            IntPtr interp
            )
        {
            //
            // NOTE: Return true if we are not aware of the Tcl interpreter;
            //       otherwise, we must return false because it is not
            //       actually owned by Eagle and we should not delete it.
            //
            lock (syncRoot)
            {
                return (tclInterps == null) ||
                    !tclInterps.ContainsValue(interp); /* O(N) */
            }
        }

        ///////////////////////////////////////////////////////////////////////

        //
        // NOTE: Used by the _Hosts.Default.WriteEngineInfo method.
        //
        internal static void AddInfo(
            bool empty,
            ref StringPairList list
            )
        {
            lock (syncRoot)
            {
                StringPairList localList = new StringPairList();

                if (empty || (interpreter != null))
                    localList.Add("Interpreter", (interpreter != null) ?
                        interpreter.ToString() : FormatOps.DisplayNull);

                if (empty || ((tclInterps != null) && (tclInterps.Count > 0)))
                    localList.Add("TclInterps", (tclInterps != null) ?
                        tclInterps.Count.ToString() : FormatOps.DisplayNull);

                if (localList.Count > 0)
                {
                    if (list == null)
                        list = new StringPairList();

                    list.Add((IPair<string>)null);
                    list.Add("Native");
                    list.Add((IPair<string>)null);
                    list.Add(localList);
                }
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Private Argument Handling Methods
        private static IntPtr StringToIntPtr(
            string text,
            ref Result error
            )
        {
            int value = 0;

            if (Value.GetInteger2(
                    text, ValueFlags.AnyInteger, null, ref value,
                    ref error) == ReturnCode.Ok)
            {
                return new IntPtr(value);
            }

            return IntPtr.Zero;
        }

        ///////////////////////////////////////////////////////////////////////

        private static ReturnCode ParseArgument(
            string arg,
            ref string protocolId,
            ref IntPtr module,
            ref IntPtr interp,
            ref bool safe,
            ref StringList list,
            ref Result error
            )
        {
            StringList localList = null;
            Result localError = null;

            //
            // NOTE: Attempt to parse the argument as a Tcl/Eagle list.  If
            //       this fails, we cannot continue.
            //
            if (Parser.SplitList(
                    null, arg, 0, Length.Invalid, false, false,
                    ref localList, ref localError) != ReturnCode.Ok)
            {
                goto done;
            }

            //
            // NOTE: Make sure the parsed list has at least the minimum number
            //       of arguments we need (i.e. the Tcl library module handle
            //       and the Tcl interpreter pointer).
            //
            if (localList.Count < MinimumArgumentCount)
            {
                localError = String.Format(
                    "not enough arguments, have {0}, need {1}",
                    localList.Count, MinimumArgumentCount);

                goto done;
            }

            //
            // NOTE: The first argument must be the protocol Id and it must
            //       match the one we were compiled with.
            //
            if (!String.Equals(localList[0], ProtocolId,
                    StringOps.SystemStringComparisonType))
            {
                localError = String.Format(
                    "protocol mismatch, have \"{0}\", need \"{1}\"",
                    localList[0], ProtocolId);

                goto done;
            }

            //
            // HACK: Assume the second argument of the list parsed from the
            //       argument actually represents the Tcl library module
            //       handle.
            //
            IntPtr localModule = StringToIntPtr(localList[1], ref localError);

            if (localModule == IntPtr.Zero)
                goto done;

            //
            // HACK: Assume the third argument of the list parsed from the
            //       argument actually represents the pointer to the Tcl
            //       interpreter.
            //
            IntPtr localInterp = StringToIntPtr(localList[2], ref localError);

            if (localInterp == IntPtr.Zero)
                goto done;

            //
            // NOTE: The fourth argument is the Tcl interpreter safety
            //       indicator.  Non-zero means the Tcl interpreter is "safe"
            //       and the Eagle interpreter MUST be as well.
            //
            bool localSafe = false;

            if (Value.GetBoolean2(localList[3], ValueFlags.AnyBoolean, true,
                    null, ref localSafe, ref localError) != ReturnCode.Ok)
            {
                goto done;
            }

            //
            // NOTE: Ok, everything was successful.  Commit changes to the
            //       parameters provided by the caller.
            //
            list = StringList.GetRange(localList, MinimumArgumentCount);

            protocolId = localList[0];
            interp = localInterp;
            module = localModule;
            safe = localSafe;

            return ReturnCode.Ok;

        done:
            error = String.Format(ParseArgumentError,
                ProtocolId, localError);

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Private Helper Methods
        private static bool HasTclInterpreters(
            ref Result error
            )
        {
            bool result = (tclInterps != null);

            if (!result)
                error = "no Tcl interpreters available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        private static string GetTclInterpreterPrefix(
            bool master,
            bool safe
            )
        {
            if (master)
                return safe ? tclNativeSafeMasterInterpPrefix :
                    tclNativeMasterInterpPrefix;
            else
                return safe ? tclNativeSafeInterpPrefix :
                    tclNativeInterpPrefix;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Public CLR Native API Integration Methods
        //
        // WARNING: This method is used to integrate with native code via the
        //          native CLR API.
        //
        public static int Startup(
            string argument /* This is the value of the "pwzArgument" argument
                             * as it was passed to native CLR API method
                             * ICLRRuntimeHost.ExecuteInDefaultAppDomain. */
            ) /* ENTRY-POINT, THREAD-SAFE */
        {
            ReturnCode code;
            string protocolId = null;
            IntPtr module = IntPtr.Zero;
            IntPtr interp = IntPtr.Zero;
            bool safe = false;
            StringList list = null;
            Result result = null;

            RuntimeOps.DebugTrace(String.Format(
                "{0}: Startup: entered, argument = {1}{2}",
                GlobalState.GetCurrentThreadId(),
                FormatOps.WrapOrNull(true, true, argument),
                Environment.NewLine), typeof(NativePackage).Name);

            code = ParseArgument(
                argument, ref protocolId, ref module, ref interp, ref safe,
                ref list, ref result);

            if (code == ReturnCode.Ok)
            {
                bool created = false;
                TclBridge tclBridge = null;

                try
                {
                    //
                    // NOTE: Lock access to the static data contained in this
                    //       class (e.g. the Eagle interpreter we use to host
                    //       the Tcl integration components).
                    //
                    lock (syncRoot)
                    {
                        //
                        // NOTE: We will need an Eagle interpreter to host the
                        //       necessary Tcl integration components;
                        //       therefore, attempt to create one now.  This
                        //       Eagle interpreter MUST be kept around for the
                        //       entire lifetime of the Tcl interpreter we are
                        //       integrating with and MUST be visible to the
                        //       CLR GC; therefore, we store it in a static
                        //       field of this class.  This is not an ideal
                        //       solution because it prevents us from having
                        //       multiple Tcl interpreters with a one-to-one
                        //       relationship to Eagle interpreters.
                        //
                        if (interpreter == null)
                        {
                            //
                            // NOTE: If we have not set the thread Id yet, do
                            //       it now.
                            //
                            if (threadId == 0)
                                threadId = GlobalState.GetCurrentThreadId();

                            //
                            // NOTE: If there are any Tcl interpreters in our
                            //       list at this point, they are stale and
                            //       should be removed.
                            //
                            if (tclInterps != null)
                                tclInterps.Clear();

                            CreateFlags createFlags =
                                Interpreter.GetStartupCreateFlags(
                                    null, CreateFlags.NativeUse,
                                    OptionOriginFlags.All, true, true);

                            if (safe)
                                createFlags |= CreateFlags.SafeAndHideUnsafe;

                            interpreter = Interpreter.Create(list, createFlags,
                                ref result);

                            created = true;

                            if (interpreter != null)
                                code = Interpreter.ProcessStartupOptions(
                                    interpreter, null, createFlags,
                                    OptionOriginFlags.All, true, true,
                                    ref result);

                            RuntimeOps.DebugTrace(String.Format(
                                "{0}: Startup: Interpreter {1}, args = {2}, " +
                                "flags = {3}, code = {4}, result = {5}{6}",
                                GlobalState.GetCurrentThreadId(),
                                (interpreter != null) ? "created" : "not created",
                                FormatOps.WrapOrNull(true, true, list),
                                FormatOps.WrapOrNull(createFlags), code,
                                FormatOps.WrapOrNull(true, true, result),
                                Environment.NewLine), typeof(NativePackage).Name);
                        }

                        //
                        // NOTE: Did we succeed at creating the Eagle
                        //       interpreter [at some point]?
                        //
                        if ((code == ReturnCode.Ok) && (interpreter != null))
                        {
                            //
                            // NOTE: Verify that the Eagle interpreter is safe
                            //       if the Tcl interpreter is safe (i.e. just
                            //       in case the Eagle interpreter was created
                            //       previously with an unsafe Tcl interpreter).
                            //
                            if (!safe || interpreter.IsSafe())
                            {
                                //
                                // NOTE: Lookup the Eagle [eval] command by
                                //       name and grab the IExecute object for
                                //       it.  This command will be the
                                //       destination for the [eagle] command in
                                //       Tcl added by this package.
                                //
                                IExecute execute = null;

                                code = interpreter.GetIExecuteViaResolvers(
                                    interpreter.GetResolveEngineFlags(true),
                                    managedCommandName,
                                    new ArgumentList(managedCommandName),
                                    true, true, ref execute, ref result);

                                if (code == ReturnCode.Ok)
                                {
                                    //
                                    // NOTE: Create the TclApi object based on
                                    //       the provided Tcl library module
                                    //       handle.  If the supplied module
                                    //       handle is invaild in some way,
                                    //       this is where the failure will
                                    //       likely be [noticed first].
                                    //
                                    ITclApi tclApi = interpreter.tclApi;

                                    if (tclApi == null)
                                    {
                                        tclApi = TclApi.Create(interpreter,
                                            String.Empty, module, ref result);

                                        RuntimeOps.DebugTrace(String.Format(
                                            "{0}: Startup: TclApi {1}, " +
                                            "module = {2}, code = {3}, " +
                                            "result = {4}{5}",
                                            GlobalState.GetCurrentThreadId(),
                                            (tclApi != null) ? "created" :
                                                "not created", module, code,
                                            FormatOps.WrapOrNull(
                                                true, true, result),
                                            Environment.NewLine),
                                            typeof(NativePackage).Name);
                                    }

                                    if (tclApi != null)
                                    {
                                        //
                                        // NOTE: Make sure the list of Tcl
                                        //       interpreters [that we should
                                        //       not delete] is initialized.
                                        //
                                        if (tclInterps == null)
                                            tclInterps = new IntPtrDictionary();

                                        //
                                        // NOTE: Make sure this Tcl interpreter
                                        //       is not already attached.
                                        //
                                        if (!tclInterps.ContainsValue(interp)) /* O(N) */
                                        {
                                            string interpName = FormatOps.Id(
                                                GetTclInterpreterPrefix(
                                                    (tclInterps.Count == 0), safe),
                                                null, interpreter.NextId());

                                            tclInterps.Add(interpName, interp);
                                            interpreter.tclInterps.Add(interpName,
                                                interp);

                                            interpreter.tclApi = tclApi;

                                            //
                                            // NOTE: Create the TclBridge object to
                                            //       translate calls from the [eagle]
                                            //       Tcl command to the [eval] Eagle
                                            //       command.
                                            //
                                            tclBridge = TclBridge.Create(interpreter,
                                                execute, ClientData.Empty, interp,
                                                nativeCommandName, true, true,
                                                ref result);

                                            RuntimeOps.DebugTrace(String.Format(
                                                "{0}: Startup: TclBridge {1}, " +
                                                "interpreter = {2}, execute = {3}, " +
                                                "interp = {4}, name = {5}, " +
                                                "code = {6}, result = {7}{8}",
                                                GlobalState.GetCurrentThreadId(),
                                                (tclBridge != null) ? "created" :
                                                    "not created",
                                                FormatOps.WrapOrNull(interpreter),
                                                FormatOps.WrapOrNull(execute), interp,
                                                FormatOps.WrapOrNull(nativeCommandName),
                                                code, FormatOps.WrapOrNull(
                                                    true, true, result),
                                                Environment.NewLine),
                                                typeof(NativePackage).Name);

                                            if (tclBridge != null)
                                            {
                                                string bridgeName =
                                                    FormatOps.TclBridgeName(
                                                        interpName, nativeCommandName);

                                                interpreter.tclBridges.Add(bridgeName,
                                                    tclBridge);
                                            }
                                            else
                                            {
                                                code = ReturnCode.Error;
                                            }
                                        }
                                        else
                                        {
                                            result = String.Format(AlreadyAttachedError,
                                                interp, FormatOps.WrapOrNull(interpreter));

                                            code = ReturnCode.Error;
                                        }
                                    }
                                    else
                                    {
                                        code = ReturnCode.Error;
                                    }
                                }
                            }
                            else
                            {
                                result = String.Format(SafeUnsafeError,
                                    interp, FormatOps.WrapOrNull(interpreter));

                                code = ReturnCode.Error;
                            }
                        }
                        else if (code != ReturnCode.Error)
                        {
                            code = ReturnCode.Error;
                        }
                    }
                }
                finally
                {
                    //
                    // NOTE: If we fail for some reason, cleanup any resources
                    //       we allocated during this method call.
                    //
                    if (code != ReturnCode.Ok)
                    {
                        if (tclBridge != null)
                        {
                            tclBridge.Dispose();
                            tclBridge = null;
                        }

                        //
                        // NOTE: Only dispose of the Eagle interpreter
                        //       if we created it during this method call;
                        //       otherwise, it may already be in use.
                        //
                        if (created && (interpreter != null))
                        {
                            interpreter.Dispose();
                            interpreter = null;
                        }
                    }
                }
            }

            //
            // NOTE: We have no way of passing the result string back to
            //       native code; therefore, just "complain" about it
            //       (e.g. to the console).
            //
            if (code != ReturnCode.Ok)
                DebugOps.Complain(interpreter, code, result);

            RuntimeOps.DebugTrace(String.Format(
                "{0}: Startup: exited, protocolId = {1}, module = {2}, " +
                "interp = {3}, safe = {4}, list = {5}, code = {6}, " +
                "result = {7}{8}", GlobalState.GetCurrentThreadId(),
                FormatOps.WrapOrNull(true, true, protocolId), module,
                interp, safe, FormatOps.WrapOrNull(true, true, list),
                code, FormatOps.WrapOrNull(true, true, result),
                Environment.NewLine), typeof(NativePackage).Name);

            return (int)code;
        }

        ///////////////////////////////////////////////////////////////////////

        //
        // WARNING: This method is used to integrate with native code via the
        //          native CLR API.
        //
        public static int Control(
            string argument /* This is the value of the "pwzArgument" argument
                             * as it was passed to native CLR API method
                             * ICLRRuntimeHost.ExecuteInDefaultAppDomain. */
            ) /* ENTRY-POINT, THREAD-SAFE */
        {
            ReturnCode code;
            string protocolId = null;
            IntPtr module = IntPtr.Zero;
            IntPtr interp = IntPtr.Zero;
            bool safe = false;
            StringList list = null;
            Result result = null;

            RuntimeOps.DebugTrace(String.Format(
                "{0}: Control: entered, argument = {1}{2}",
                GlobalState.GetCurrentThreadId(),
                FormatOps.WrapOrNull(true, true, argument),
                Environment.NewLine), typeof(NativePackage).Name);

            code = ParseArgument(
                argument, ref protocolId, ref module, ref interp, ref safe,
                ref list, ref result);

            if (code == ReturnCode.Ok)
            {
                // do nothing.
            }

            //
            // NOTE: We have no way of passing the result string back to
            //       native code; therefore, just "complain" about it
            //       (e.g. to the console).
            //
            if (code != ReturnCode.Ok)
                DebugOps.Complain(interpreter, code, result);

            RuntimeOps.DebugTrace(String.Format(
                "{0}: Control: exited, protocolId = {1}, module = {2}, " +
                "interp = {3}, safe = {4}, list = {5}, code = {6}, " +
                "result = {7}{8}", GlobalState.GetCurrentThreadId(),
                FormatOps.WrapOrNull(true, true, protocolId), module,
                interp, safe, FormatOps.WrapOrNull(true, true, list),
                code, FormatOps.WrapOrNull(true, true, result),
                Environment.NewLine), typeof(NativePackage).Name);

            return (int)code;
        }

        ///////////////////////////////////////////////////////////////////////

        //
        // WARNING: This method is used to integrate with native code via the
        //          native CLR API.
        //
        public static int Detach(
            string argument /* This is the value of the "pwzArgument" argument
                             * as it was passed to native CLR API method
                             * ICLRRuntimeHost.ExecuteInDefaultAppDomain. */
            ) /* ENTRY-POINT, THREAD-SAFE */
        {
            ReturnCode code;
            string protocolId = null;
            IntPtr module = IntPtr.Zero;
            IntPtr interp = IntPtr.Zero;
            bool safe = false;
            StringList list = null;
            Result result = null;

            RuntimeOps.DebugTrace(String.Format(
                "{0}: Detach: entered, argument = {1}{2}",
                GlobalState.GetCurrentThreadId(),
                FormatOps.WrapOrNull(true, true, argument),
                Environment.NewLine), typeof(NativePackage).Name);

            code = ParseArgument(
                argument, ref protocolId, ref module, ref interp, ref safe,
                ref list, ref result);

            if (code == ReturnCode.Ok)
            {
                lock (syncRoot)
                {
                    if (interpreter != null)
                    {
                        try
                        {
                            code = interpreter.DisposeTclBridges(
                                interp, false, ref result);

                            if (code == ReturnCode.Ok)
                            {
                                //
                                // NOTE: Now that we are sure that the matching
                                //       TclBridge objects in the Eagle
                                //       interpreter has been disposed, remove
                                //       this Tcl interpreter from the list
                                //       [that we should not delete] because
                                //       this is only an issue during disposal
                                //       of the Eagle interpreter.
                                //
                                if (tclInterps != null)
                                {
                                    tclInterps.Remove(interp); /* O(N) */

                                    if (tclInterps.Count == 0)
                                        tclInterps = null;
                                }

                                //
                                // NOTE: If we are no longer attached to any
                                //       Tcl interpreters, zero our thread Id
                                //       to zero in case we need to be
                                //       re-initialized on a different thread.
                                //
                                if (tclInterps == null)
                                    threadId = 0;
                            }
                        }
                        catch (Exception e)
                        {
                            result = e;
                            code = ReturnCode.Error;
                        }
                    }
                }
            }

            //
            // NOTE: We have no way of passing the result string back to
            //       native code; therefore, just "complain" about it
            //       (e.g. to the console).
            //
            if (code != ReturnCode.Ok)
                DebugOps.Complain(interpreter, code, result);

            RuntimeOps.DebugTrace(String.Format(
                "{0}: Detach: exited, protocolId = {1}, module = {2}, " +
                "interp = {3}, safe = {4}, list = {5}, code = {6}, " +
                "result = {7}{8}", GlobalState.GetCurrentThreadId(),
                FormatOps.WrapOrNull(true, true, protocolId), module,
                interp, safe, FormatOps.WrapOrNull(true, true, list),
                code, FormatOps.WrapOrNull(true, true, result),
                Environment.NewLine), typeof(NativePackage).Name);

            return (int)code;
        }

        ///////////////////////////////////////////////////////////////////////

        //
        // WARNING: This method is used to integrate with native code via the
        //          native CLR API.
        //
        public static int Shutdown(
            string argument /* This is the value of the "pwzArgument" argument
                             * as it was passed to native CLR API method
                             * ICLRRuntimeHost.ExecuteInDefaultAppDomain. */
            ) /* ENTRY-POINT, THREAD-SAFE */
        {
            ReturnCode code;
            string protocolId = null;
            IntPtr module = IntPtr.Zero;
            IntPtr interp = IntPtr.Zero;
            bool safe = false;
            StringList list = null;
            Result result = null;

            RuntimeOps.DebugTrace(String.Format(
                "{0}: Shutdown: entered, argument = {1}{2}",
                GlobalState.GetCurrentThreadId(),
                FormatOps.WrapOrNull(true, true, argument),
                Environment.NewLine), typeof(NativePackage).Name);

            code = ParseArgument(
                argument, ref protocolId, ref module, ref interp, ref safe,
                ref list, ref result);

            if (code == ReturnCode.Ok)
            {
                lock (syncRoot)
                {
                    if (interpreter != null)
                    {
                        try
                        {
                            //
                            // NOTE: Attempt to dispose of the Eagle interpreter
                            //       now.  In theory, this may throw exceptions;
                            //       however, since it should not attempt to
                            //       unload Tcl, so that is unlikely.
                            //
                            interpreter.Dispose(); /* throw */
                            interpreter = null;

                            //
                            // NOTE: Now that we are sure the Eagle interpreter
                            //       has been disposed, remove all the Tcl
                            //       interpreters from the list [that we should
                            //       not delete] because this is only an issue
                            //       during disposal of the Eagle interpreter.
                            //
                            if (tclInterps != null)
                            {
                                tclInterps.Clear();
                                tclInterps = null;
                            }

                            //
                            // NOTE: Zero out our thread Id in case we need to
                            //       be re-initialized on a different thread.
                            //
                            threadId = 0;
                        }
                        catch (Exception e)
                        {
                            result = e;
                            code = ReturnCode.Error;
                        }
                    }
                }
            }

            //
            // NOTE: We have no way of passing the result string back to
            //       native code; therefore, just "complain" about it
            //       (e.g. to the console).
            //
            if (code != ReturnCode.Ok)
                DebugOps.Complain(interpreter, code, result);

            RuntimeOps.DebugTrace(String.Format(
                "{0}: Shutdown: exited, protocolId = {1}, module = {2}, " +
                "interp = {3}, safe = {4}, list = {5}, code = {6}, " +
                "result = {7}{8}", GlobalState.GetCurrentThreadId(),
                FormatOps.WrapOrNull(true, true, protocolId), module,
                interp, safe, FormatOps.WrapOrNull(true, true, list),
                code, FormatOps.WrapOrNull(true, true, result),
                Environment.NewLine), typeof(NativePackage).Name);

            return (int)code;
        }
        #endregion
    }
}
